from typing import Optional, List
import json
from tenacity import retry, stop_after_attempt
from langchain_core.output_parsers import JsonOutputParser
from langchain_core.prompts import PromptTemplate
from langchain_openai import ChatOpenAI
from pydantic import BaseModel, Field

# LLM prompt for normalizing headers
SPLITTER_PROMPT = "Given the following headers with incorrect levels, adjust them to the correct hierarchical structure. Do not rearrange the headers. Output results exclusively in json format.\n{format_instructions}\n{query}\n"


# Header formats for llm
class HeaderFormat(BaseModel):
    text: str = Field(description="header text")
    level: int = Field(description="header level")


class HeadersFormat(BaseModel):
    headers: List[HeaderFormat] = Field(description="array of headers")


# Class for normalizing markdown headers
class HeaderNormalizer:
    def __init__(
        self,
        max_heading_level: int = 3,
        llm_for_headings: bool = False,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        model: Optional[str] = "gpt-4o",
    ):
        self.llm_for_headings = llm_for_headings
        self.base_url = base_url
        self.api_key = api_key
        self.model = model
        self.max_heading_level = max_heading_level

    # Use LLM to normalize markdown headers
    @retry(stop=stop_after_attempt(3))
    def _normalize_headers_with_llm(self, headers: dict) -> dict:
        client = ChatOpenAI(
            openai_api_base=self.base_url,
            openai_api_key=self.api_key,
            model_name=self.model,
        )
        parser = JsonOutputParser(pydantic_object=HeadersFormat)
        prompt = PromptTemplate(
            template=SPLITTER_PROMPT,
            input_variables=["query"],
            partial_variables={"format_instructions": parser.get_format_instructions()},
        )
        proc_chain = prompt | client | parser
        corrected_headers = proc_chain.invoke(
            {"query": json.dumps(headers, ensure_ascii=False)}
        )
        assert len(headers) == len(corrected_headers)
        return corrected_headers

    # Normalize markdown headers manually
    def _normalize_headers_simple(self, headers: dict) -> dict:
        corrected_headers = [
            {"text": x["text"], "level": min(x["level"], self.max_heading_level)}
            for x in headers["headers"]
        ]
        return {"headers": corrected_headers}

    # Normalize markdown headers
    def __call__(self, text: str) -> str:
        # Normalize headers
        headers_texts = [
            line for line in text.split("\n") if len(line) > 0 and line[0] == "#"
        ]
        headers = [
            {"text": x.replace("#", "").strip(), "level": x.count("#")}
            for x in headers_texts
        ]
        headers = {"headers": headers}
        # Try different tec techniques
        corrected_headers = None
        if self.llm_for_headings:
            try:
                corrected_headers = self._normalize_headers_with_llm(headers)
            except Exception as e:
                print(f"Failed normalizing headers: {e}")
        if not corrected_headers:
            corrected_headers = self._normalize_headers_simple(headers)
        # Replace headers in text
        correct_texts = [
            f'{"#"*x["level"]} {x["text"]}' for x in corrected_headers["headers"]
        ]
        for i in range(len(headers_texts)):
            text = text.replace(f"\n{headers_texts[i]}\n", f"\n{correct_texts[i]}\n")
        return text



from typing import Optional, Union, Literal, List
import os
from pathlib import Path
import hashlib
import json
import shutil
from dataclasses import asdict
from .models import RAGDocument


# Abstraction for managing cache
class CacheManager:
    def __init__(self, cache_dir: Optional[str] = None):
        self.cache_dir = cache_dir
        if cache_dir:
            os.makedirs(cache_dir, exist_ok=True)
            self.cache_dir = Path(cache_dir)
            os.makedirs(self.cache_dir / "convert", exist_ok=True)
            os.makedirs(self.cache_dir / "process", exist_ok=True)
            os.makedirs(self.cache_dir / "rag", exist_ok=True)

    # Get text checksum
    def _get_text_hash(self, text: str) -> str:
        text_hash = hashlib.blake2b(digest_size=32)
        text_hash.update(text.encode("utf8"))
        return text_hash.hexdigest()

    # Get file checksum
    def _get_file_hash(self, path: str) -> str:
        with open(path, "rb") as f:
            file_hash = hashlib.blake2b(digest_size=32)
            while chunk := f.read(8192):
                file_hash.update(chunk)
        return file_hash.hexdigest()

    # Load from cache
    def load(
        self,
        cache_type: Literal["convert", "process", "rag"],
        *args,
        path: Optional[str] = None,
        text: Optional[str] = None,
    ) -> Union[str, dict, None]:
        if not self.cache_dir or not (path or text):
            return None
        filename = self._get_file_hash(path) if path else self._get_text_hash(text)
        format = "json" if cache_type == "rag" else "md"
        cache_path = self.cache_dir / cache_type / f"{filename}.{format}"
        if os.path.exists(cache_path):
            with open(cache_path, "r") as f:
                output = f.read()
            if format == "json":
                data = json.loads(output)
                output = [RAGDocument(**x) for x in data]
            return output

    # Save to cache
    def save(
        self,
        cache_type: Literal["convert", "process", "rag"],
        *args,
        path: Optional[str] = None,
        text: Optional[str] = None,
        data: Union[str, List[RAGDocument]],
    ) -> None:
        if not self.cache_dir or not (path or text):
            return None
        filename = self._get_file_hash(path) if path else self._get_text_hash(text)
        format = "json" if cache_type == "rag" else "md"
        cache_path = self.cache_dir / cache_type / f"{filename}.{format}"
        with open(cache_path, "w") as f:
            if format == "json":
                new_data = [asdict(x) for x in data]
                f.write(json.dumps(new_data))
            else:
                f.write(data)

    # Clean cache
    def clean(self, cache_type: Optional[Literal["convert", "process", "rag"]] = None):
        if cache_type:
            shutil.rmtree(self.cache_dir / cache_type)
            os.makedirs(self.cache_dir / cache_type, exist_ok=True)
        else:
            shutil.rmtree(self.cache_dir)
            os.makedirs(self.cache_dir)



from typing import Literal, Optional, List, Any
from jinja2 import Template, Environment, PackageLoader
from .models import RAGDocument

# Define constants
DEFAULT_TEMPLATE = Environment(
    loader=PackageLoader("raggen", "templates")
).get_template("meta_embed")


# Abstraction for dealing with metadata related operations
class MetadataManager:
    def __init__(
        self,
        encoder: Optional[Any] = None,
        template: Optional[str] = None,
        custom_meta_placement: Literal["before", "after"] = "before",
        field_names: Optional[dict] = None,
    ):
        self.encoder = encoder
        self.template = Template(template) if template else DEFAULT_TEMPLATE
        self.custom_meta_placement = custom_meta_placement
        self.field_names = (
            {"section": "Section", "subsection": "Subsection", "paragraph": "Paragraph"}
            if not field_names
            else field_names
        )

    # Embed metadata to text fragments
    def embed_metadata(self, documents: List[RAGDocument]) -> List[RAGDocument]:
        for doc in documents:
            if not doc.metadata:
                continue
            metadata = {
                self.field_names[key] if key in self.field_names else key: value
                for key, value in doc.metadata.items()
            }
            new_text = self.template.render(text=doc.text, metadata=metadata)
            doc.text = new_text.strip()
            if doc.length != None and self.encoder:
                doc.length = len(self.encoder.encode(doc.text))
        return documents

    # Add custom metadata
    def add_metadata(self, documents: List[RAGDocument], metadata: dict) -> List[RAGDocument]:
        for doc in documents:
            if self.custom_meta_placement == "before":
                new_meta = metadata.copy()
                new_meta.update(doc.metadata)
                doc.metadata = new_meta
            else:
                doc.metadata.update(metadata)
        return documents



from langchain_text_splitters import (
    MarkdownHeaderTextSplitter,
    RecursiveCharacterTextSplitter,
)
from typing import List, Tuple
from .models import RAGDocument
import tiktoken
import re

# Table finder reges
TABLE_PATTERN = re.compile(r"(\n\n(\|.+\|\n)+\n)")


# Markdown text splitter
class MarkdownSplitter:
    # Initialization
    def __init__(
        self,
        chunk_size: int = 256,
        chunk_overlap: int = 30,
        preserve_tables: bool = True,
    ) -> None:
        headers_to_split = [
            ("#", "section"),
            ("##", "subsection"),
            ("###", "paragraph"),
        ]
        self.preserve_tables = preserve_tables
        self.markdown_splitter = MarkdownHeaderTextSplitter(
            headers_to_split_on=headers_to_split, strip_headers=True
        )
        self.encoder = tiktoken.encoding_for_model("gpt-4o")
        self.text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            length_function=lambda text: len(self.encoder.encode(text)),
        )

    # Extract tables and replace them with placeholders
    def extract_tables(self, text: str) -> Tuple[str, List[str]]:
        tables = TABLE_PATTERN.findall(text)
        tables = [x[0].strip() for x in tables]
        for i, table in enumerate(tables):
            text = text.replace(table, f"{{table_{i}}}")
        return text, tables

    # Restore extracted tables in all splits
    def _restore_tables(self, splits: List[RAGDocument], tables: List[str]):
        table_tags = [f"{{table_{i}}}" for i in range(len(tables))]
        for split in splits:
            for i, tag in enumerate(table_tags):
                if tag in split.text:
                    split.text = split.text.replace(tag, f"\n{tables[i]}\n")

    # Split markdown document
    def __call__(self, text: str) -> List[RAGDocument]:
        if self.preserve_tables:
            text, tables = self.extract_tables(text)
        md_header_splits = self.markdown_splitter.split_text(text)
        splits = self.text_splitter.split_documents(md_header_splits)
        splits = [
            RAGDocument(
                text=x.page_content,
                metadata=x.metadata if x.metadata else None,
                length=len(self.encoder.encode(x.page_content)),
            )
            for x in splits
        ]
        if self.preserve_tables:
            self._restore_tables(splits, tables)
        return splits
